;;; guile-irrlicht --- FFI bindings for Irrlicht Engine
;;; Copyright (C) 2020 Javier Sancho <jsf@jsancho.org>
;;;
;;; This file is part of guile-irrlicht.
;;;
;;; Guile-irrlicht is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU Lesser General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Guile-irrlicht is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with guile-irrlicht.  If not, see
;;; <http://www.gnu.org/licenses/>.


(define-module (irrlicht gui)
  #:use-module (oop goops)
  #:use-module (ice-9 optargs)
  #:use-module (irrlicht base)
  #:use-module (irrlicht foreign)
  #:use-module ((irrlicht io) #:select (<attribute-exchanging-object>))
  #:use-module ((irrlicht irr) #:select (<event-receiver> <reference-counted>)))


;; ICursorControl
(define-class <cursor-control> (<reference-counted>)
  (irr-class #:init-value "ICursorControl"))

(define-method (set-visible! (cursor-control <cursor-control>) visible)
  (let ((setVisible (get-irrlicht-proc "setVisible" cursor-control)))
    (setVisible cursor-control visible)))

(export <cursor-control> set-visible!)


;; IGUIElement
(define-class <gui-element> (<attribute-exchanging-object> <event-receiver>)
  (irr-class #:init-value "IGUIElement"))

(export <gui-element>)


;; IGUIEnvironment
(define-class <gui-environment> (<reference-counted>)
  (irr-class #:init-value "IGUIEnvironment"))

(define-method (add-button! (gui-environment <gui-environment>) rectangle . rest)
  (let-keywords rest #f
        ((parent (make <gui-element>))
         (id -1)
         (text "")
         (tooltiptext ""))
    (let ((addButton (get-irrlicht-proc "addButton" gui-environment parent)))
      (addButton gui-environment rectangle parent id text tooltiptext))))

(define-method (add-image! (gui-environment <gui-environment>) image pos . rest)
  (let-keywords rest #f
        ((use-alpha-channel #t)
         (parent (make <gui-element>))
         (id -1)
         (text ""))
    (let ((addImage (get-irrlicht-proc "addImage" gui-environment parent)))
      (addImage gui-environment image pos use-alpha-channel parent id text))))

(define-method (add-listbox! (gui-environment <gui-environment>) rectangle . rest)
  (let-keywords rest #f
        ((parent (make <gui-element>))
         (id -1)
         (draw-background #f))
    (let ((addListBox (get-irrlicht-proc "addListBox" gui-environment parent)))
      (addListBox gui-environment rectangle parent id draw-background))))

(define-method (add-scrollbar! (gui-environment <gui-environment>) horizontal rectangle . rest)
  (let-keywords rest #f
        ((parent (make <gui-element>))
         (id -1))
    (let ((addScrollBar (get-irrlicht-proc "addScrollBar" gui-environment parent)))
      (addScrollBar gui-environment horizontal rectangle parent id))))

(define-method (add-static-text! (gui-environment <gui-environment>) text rectangle . rest)
  (let-keywords rest #f
        ((border #f)
         (word-wrap #t)
         (parent (make <gui-element>))
         (id -1)
         (fill-background #f))
    (let ((addStaticText (get-irrlicht-proc "addStaticText" gui-environment parent)))
      (addStaticText gui-environment text rectangle border word-wrap parent
                     id fill-background))))

(define-method (draw-all (gui-environment <gui-environment>))
  ((get-irrlicht-proc "drawAll" gui-environment)
   gui-environment))

(define-method (get-built-in-font (gui-environment <gui-environment>))
  (let ((getBuiltInFont (get-irrlicht-proc "getBuiltInFont" gui-environment)))
    (getBuiltInFont gui-environment)))

(define-method (get-font (gui-environment <gui-environment>) filename)
  (let ((getFont (get-irrlicht-proc "getFont" gui-environment)))
    (getFont gui-environment filename)))

(define-method (get-skin (gui-environment <gui-environment>))
  (let ((getSkin (get-irrlicht-proc "getSkin" gui-environment)))
    (getSkin gui-environment)))

(export <gui-environment> add-button! add-image! add-listbox! add-scrollbar! add-static-text!
        draw-all get-built-in-font get-font get-skin)


;; IGUIStaticText
(define-class <gui-static-text> (<gui-element>)
  (irr-class #:init-value "IGUIStaticText"))

(define-method (set-override-color! (static-text <gui-static-text>) color)
  (let ((setOverrideColor (get-irrlicht-proc "setOverrideColor" static-text)))
    (setOverrideColor static-text color)))

(export <gui-static-text> set-override-color!)


;; IGUIImage
(define-class <gui-image> (<gui-element>)
  (irr-class #:init-value "IGUIImage"))

(export <gui-image>)


;; IGUISkin
(define-class <gui-skin> (<attribute-exchanging-object>)
  (irr-class #:init-value "IGUISkin"))

(define-method (get-color (skin <gui-skin>) color)
  (let ((getColor (get-irrlicht-proc "getColor" skin)))
    (getColor skin color)))

(define-method (set-font! (skin <gui-skin>) font . rest)
  (let-keywords rest #f
        ((which 'default))
    (let ((setFont (get-irrlicht-proc "setFont" skin)))
      (setFont skin font which))))

(define-method (set-color! (skin <gui-skin>) which new-color)
  (let ((setColor (get-irrlicht-proc "setColor" skin)))
    (setColor skin which new-color)))

(export <gui-skin> get-color set-font! set-color!)


;; IGUIFont
(define-class <gui-font> (<reference-counted>)
  (irr-class #:init-value "IGUIFont"))

(export <gui-font>)


;; IGUIButton
(define-class <gui-button> (<gui-element>)
  (irr-class #:init-value "IGUIButton"))

(export <gui-button>)


;; IGUIScrollBar
(define-class <gui-scrollbar> (<gui-element>)
  (irr-class #:init-value "IGUIScrollBar"))

(define-method (get-position (scrollbar <gui-scrollbar>))
  (let ((getPos (get-irrlicht-proc "getPos" scrollbar)))
    (getPos scrollbar)))

(define-method (set-max! (scrollbar <gui-scrollbar>) max)
  (let ((setMax (get-irrlicht-proc "setMax" scrollbar)))
    (setMax scrollbar max)))

(define-method (set-position! (scrollbar <gui-scrollbar>) pos)
  (let ((setPos (get-irrlicht-proc "setPos" scrollbar)))
    (setPos scrollbar pos)))

(export <gui-scrollbar> get-position set-max! set-position!)


;; IGUIListBox
(define-class <gui-listbox> (<gui-element>)
  (irr-class #:init-value "IGUIListBox"))

(export <gui-listbox>)
