;;; guile-irrlicht --- FFI bindings for Irrlicht Engine
;;; Copyright (C) 2020 Javier Sancho <jsf@jsancho.org>
;;;
;;; This file is part of guile-irrlicht.
;;;
;;; Guile-irrlicht is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU Lesser General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Guile-irrlicht is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with guile-irrlicht.  If not, see
;;; <http://www.gnu.org/licenses/>.


(define-module (irrlicht scene)
  #:use-module (oop goops)
  #:use-module (ice-9 optargs)
  #:use-module (irrlicht base)
  #:use-module (irrlicht foreign)
  #:use-module ((irrlicht io) #:select (<attribute-exchanging-object>))
  #:use-module ((irrlicht irr) #:select (<event-receiver> <key-map> <reference-counted>))
  #:use-module ((irrlicht video) #:select (<texture>)))


;; IMesh
(define-class <mesh> (<reference-counted>)
  (irr-class #:init-value "IMesh"))

(export <mesh>)


;; IAnimatedMesh
(define-class <animated-mesh> (<mesh>)
  (irr-class #:init-value "IAnimatedMesh"))

(export <animated-mesh>)


;; ISceneNodeAnimator
(define-class <scene-node-animator> (<attribute-exchanging-object> <event-receiver>)
  (irr-class #:init-value "ISceneNodeAnimator"))

(export <scene-node-animator>)


;; ISceneManager
(define-class <scene-manager> (<reference-counted>)
  (irr-class #:init-value "ISceneManager"))

(define-method (add-animated-mesh-scene-node! (scene-manager <scene-manager>)
                                              (mesh <animated-mesh>)
                                              . rest)
  (let-keywords rest #f
        ((parent (make <scene-node>))
         (id -1)
         (position '(0 0 0))
         (rotation '(0 0 0))
         (scale '(1 1 1))
         (also-add-if-mesh-pointer-zero #f))
    (let* ((addAnimatedMeshSceneNode (get-irrlicht-proc "addAnimatedMeshSceneNode"
                                                        scene-manager
                                                        parent))
           (node (addAnimatedMeshSceneNode scene-manager mesh parent id position
                                           rotation scale also-add-if-mesh-pointer-zero)))
      (cond ((null-object? node)
             (error "In procedure add-animated-mesh-scene-node!: Scene node cannot be created"))
            (else
             node)))))

(define-method (add-camera-scene-node! (scene-manager <scene-manager>) . rest)
  (let-keywords rest #f
        ((parent (make <scene-node>))
         (position '(0 0 0))
         (lookat '(0 0 100))
         (id -1)
         (make-active #t))
    (let ((addCameraSceneNode (get-irrlicht-proc "addCameraSceneNode" scene-manager parent)))
      (addCameraSceneNode scene-manager parent position lookat id make-active))))

(define-method (add-camera-scene-node-fps! (scene-manager <scene-manager>) . rest)
  (let-keywords rest #f
        ((parent (make <scene-node>))
         (rotate-speed 100)
         (move-speed 0.5)
         (id -1)
         (key-map-array (make <key-map>))
         (key-map-size 0)
         (no-vertical-movement #f)
         (jump-speed 0)
         (invert-mouse #f)
         (make-active #t))
    (let ((addCameraSceneNodeFPS (get-irrlicht-proc "addCameraSceneNodeFPS"
                                                    scene-manager parent)))
      (addCameraSceneNodeFPS scene-manager #:parent parent #:rotate-speed rotate-speed
                             #:move-speed move-speed #:id id #:key-map-array key-map-array
                             #:key-map-size key-map-size
                             #:no-vertical-movement no-vertical-movement
                             #:jump-speed jump-speed #:invert-mouse invert-mouse
                             #:make-active make-active))))

(define-method (add-cube-scene-node! (scene-manager <scene-manager>) . rest)
  (let-keywords rest #f
        ((size 10)
         (parent (make <scene-node>))
         (id -1)
         (position '(0 0 0))
         (rotation '(0 0 0))
         (scale '(1 1 1)))
    (let ((addCubeSceneNode (get-irrlicht-proc "addCubeSceneNode" scene-manager parent)))
      (addCubeSceneNode scene-manager size parent id position rotation scale))))

(define-method (add-custom-scene-node! (scene-manager <scene-manager>) proc-render
                                       proc-get-bounding-box  proc-get-material-count
                                       proc-get-material . rest)
  (let-keywords rest #f
        ((parent (make <scene-node>))
         (id -1)
         (position '(0 0 0))
         (rotation '(0 0 0))
         (scale '(1 1 1)))
    (let ((addCustomSceneNode (get-irrlicht-proc "addCustomSceneNode" scene-manager parent)))
      (addCustomSceneNode scene-manager proc-render proc-get-bounding-box proc-get-material-count
                          proc-get-material parent id position rotation scale))))

(define-method (add-octree-scene-node! (scene-manager <scene-manager>)
                                       (mesh <animated-mesh>)
                                       . rest)
  (let-keywords rest #f
        ((parent (make <scene-node>))
         (id -1)
         (minimal-polys-per-node 512)
         (also-add-if-mesh-pointer-zero #f))
    (let ((addOctreeSceneNode (get-irrlicht-proc "addOctreeSceneNode" scene-manager parent mesh)))
      (addOctreeSceneNode scene-manager mesh parent id minimal-polys-per-node
                          also-add-if-mesh-pointer-zero))))

(define-method (add-octree-scene-node! (scene-manager <scene-manager>)
                                       (mesh <mesh>)
                                       . rest)
  (let-keywords rest #f
        ((parent (make <scene-node>))
         (id -1)
         (minimal-polys-per-node 256)
         (also-add-if-mesh-pointer-zero #f))
    (let ((addOctreeSceneNode (get-irrlicht-proc "addOctreeSceneNode" scene-manager parent mesh)))
      (addOctreeSceneNode scene-manager mesh parent id minimal-polys-per-node
                          also-add-if-mesh-pointer-zero))))

(define-method (add-sphere-scene-node! (scene-manager <scene-manager>) . rest)
  (let-keywords rest #f
        ((radius 5)
         (poly-count 16)
         (parent (make <scene-node>))
         (id -1)
         (position '(0 0 0))
         (rotation '(0 0 0))
         (scale '(1 1 1)))
    (let ((addSphereSceneNode (get-irrlicht-proc "addSphereSceneNode" scene-manager parent)))
      (addSphereSceneNode scene-manager radius poly-count parent id position rotation scale))))

(define-method (create-fly-circle-animator (scene-manager <scene-manager>) . rest)
  (let-keywords rest #f
        ((center '(0 0 0))
         (radius 100)
         (speed 0.001)
         (direction '(0 1 0))
         (start-position 0)
         (radius-ellipsoid 0))
    (let ((createFlyCircleAnimator (get-irrlicht-proc "createFlyCircleAnimator" scene-manager)))
      (createFlyCircleAnimator scene-manager center radius speed direction start-position
                               radius-ellipsoid))))

(define-method (create-fly-straight-animator (scene-manager <scene-manager>)
                                             start-point end-point time-for-way . rest)
  (let-keywords rest #f
        ((loop #f)
         (pingpong #f))
    (let ((createFlyStraightAnimator (get-irrlicht-proc "createFlyStraightAnimator" scene-manager)))
      (createFlyStraightAnimator scene-manager start-point end-point time-for-way
                                 loop pingpong))))

(define-method (create-rotation-animator (scene-manager <scene-manager>) rotation-speed)
  (let ((createRotationAnimator (get-irrlicht-proc "createRotationAnimator" scene-manager)))
    (createRotationAnimator scene-manager rotation-speed)))

(define-method (draw-all (scene-manager <scene-manager>))
  ((get-irrlicht-proc "drawAll" scene-manager)
   scene-manager))

(define-method (get-mesh (scene-manager <scene-manager>) filename)
  (let* ((getMesh (get-irrlicht-proc "getMesh" scene-manager))
         (mesh (getMesh scene-manager filename)))
    (cond ((null-object? mesh)
           (error "In procedure get-mesh: Mesh cannot be created"))
          (else
           mesh))))

(define-method (get-root-scene-node (scene-manager <scene-manager>))
  (let ((getRootSceneNode (get-irrlicht-proc "getRootSceneNode" scene-manager)))
    (getRootSceneNode scene-manager)))

(export <scene-manager> add-animated-mesh-scene-node! add-cube-scene-node! add-camera-scene-node!
        add-camera-scene-node-fps! add-custom-scene-node! add-octree-scene-node!
        add-sphere-scene-node! create-fly-circle-animator create-fly-straight-animator
        create-rotation-animator draw-all get-root-scene-node get-mesh)


;; ISceneNode
(define-class <scene-node> (<attribute-exchanging-object>)
  (irr-class #:init-value "ISceneNode"))

(define-method (add-animator! (node <scene-node>) (animator <scene-node-animator>))
  (let ((addAnimator (get-irrlicht-proc "addAnimator" node)))
    (addAnimator node animator)))

(define-method (get-absolute-transformation (node <scene-node>))
  (let ((getAbsoluteTransformation (get-irrlicht-proc "getAbsoluteTransformation" node)))
    (getAbsoluteTransformation node)))

(define-method (get-position (node <scene-node>))
  (let ((getPosition (get-irrlicht-proc "getPosition" node)))
    (getPosition node)))

(define-method (set-material-flag! (node <scene-node>) flag new-value)
  ((get-irrlicht-proc "setMaterialFlag" node)
   node
   flag
   new-value))

(define-method (set-material-texture! (node <scene-node>) texture-layer (texture <texture>))
  ((get-irrlicht-proc "setMaterialTexture" node)
   node
   texture-layer
   texture))

(define-method (set-position! (node <scene-node>) new-pos)
  (let ((setPosition (get-irrlicht-proc "setPosition" node)))
    (setPosition node new-pos)))

(define-method (set-rotation! (node <scene-node>) rotation)
  (let ((setRotation (get-irrlicht-proc "setRotation" node)))
    (setRotation node rotation)))

(define-method (set-scale! (node <scene-node>) scale)
  (let ((setScale (get-irrlicht-proc "setScale" node)))
    (setScale node scale)))

(export <scene-node> add-animator! get-absolute-transformation get-position set-material-flag!
        set-material-texture! set-position! set-rotation! set-scale!)


;; IAnimatedMeshSceneNode
(define-class <animated-mesh-scene-node> (<scene-node>)
  (irr-class #:init-value "IAnimatedMeshSceneNode"))

(define-method (set-animation-speed! (node <animated-mesh-scene-node>) frames-per-second)
  (let ((setAnimationSpeed (get-irrlicht-proc "setAnimationSpeed" node)))
    (setAnimationSpeed node frames-per-second)))

(define-method (set-frame-loop! (node <animated-mesh-scene-node>) begin end)
  (let ((setFrameLoop (get-irrlicht-proc "setFrameLoop" node)))
    (setFrameLoop node begin end)))

(define-method (set-md2-animation! (node <animated-mesh-scene-node>) anim)
  ((get-irrlicht-proc "setMD2Animation" node)
   node
   anim))

(export <animated-mesh-scene-node> set-animation-speed! set-frame-loop! set-md2-animation!)


;; ICameraSceneNode
(define-class <camera-scene-node> (<scene-node>)
  (irr-class #:init-value "ICameraSceneNode"))

(export <camera-scene-node>)


;; IMeshSceneNode
(define-class <mesh-scene-node> (<scene-node>)
  (irr-class #:init-value "IMeshSceneNode"))

(export <mesh-scene-node>)
